import copy
import os
from typing import Any

import tomli_w
from qcio import CalcType, ProgramInput

from qccodec.exceptions import EncoderError
from qccodec.models import NativeInput

SUPPORTED_CALCTYPES = {
    CalcType.conformer_search,
    CalcType.optimization,
    CalcType.energy,
    CalcType.gradient,
    CalcType.hessian,
}


def encode(inp_obj: ProgramInput) -> NativeInput:
    """Translate a ProgramInput into CREST inputs files.

    Args:
        inp_obj: The qcio ProgramInput object for a computation.

    Returns:
        NativeInput with .input_files being a crest.toml file and .geometry_file the
            Structure's xyz file.
    """
    validate_input(inp_obj)
    struct_filename = "structure.xyz"

    return NativeInput(
        input_file=tomli_w.dumps(_to_toml_dict(inp_obj, struct_filename)),
        geometry_file=inp_obj.structure.to_xyz(),
        geometry_filename=struct_filename,
    )


def validate_input(inp_obj: ProgramInput):
    """Validate the input for CREST.

    Args:
        inp_obj: The qcio ProgramInput object for a computation.

    Raises:
        EncoderError: If the input is invalid.
    """
    # These values come from other parts of the ProgramInput and should not be set
    # in the keywords.
    non_allowed_keywords = ["charge", "uhf"]
    for keyword in non_allowed_keywords:
        if keyword in inp_obj.keywords:
            raise EncoderError(
                f"{keyword} should not be set in keywords for CREST. It is already set "
                "on the Structure or ProgramInput elsewhere.",
            )
    if "runtype" in inp_obj.keywords:
        _validate_runtype_calctype(inp_obj.keywords["runtype"], inp_obj.calctype)


def _validate_runtype_calctype(runtype: str, calctype: CalcType):
    """Validate that the runtype is supported for the calctype."""
    invalid_runtype = False
    valid_runtypes = set()

    if calctype == CalcType.conformer_search:
        valid_runtypes = {"imtd-gc", "imtd-smtd", "entropy", "nci", "nci-mtd"}
        if runtype not in valid_runtypes:
            invalid_runtype = True

    elif calctype == CalcType.optimization:
        valid_runtypes = {"optimize", "ancopt"}
        if runtype not in valid_runtypes:
            invalid_runtype = True

    elif calctype in {CalcType.energy, CalcType.gradient}:
        valid_runtypes = {"singlepoint"}
        if runtype not in valid_runtypes:
            invalid_runtype = True

    elif calctype == CalcType.hessian:
        valid_runtypes = {"numhess"}
        if runtype not in valid_runtypes:
            invalid_runtype = True

    if invalid_runtype:
        raise EncoderError(
            f"Unsupported runtype {runtype} for calctype {calctype}. Valid runtypes "
            f"are: {valid_runtypes}.",
        )


def _to_toml_dict(inp_obj: ProgramInput, struct_filename: str) -> dict[str, Any]:
    """Convert a ProgramInput object to a dictionary in the CREST format of TOML.

    This function makes it easier to test for the correct TOML structure.
    """
    # Start with existing keywords
    toml_dict = copy.deepcopy(inp_obj.keywords)

    # Top level keywords
    # Logical cores was 10% faster than physical cores, so not using psutil
    toml_dict.setdefault("threads", min(os.cpu_count() or 16, 16))
    toml_dict["input"] = struct_filename

    # Set default runtype if not already set
    if "runtype" not in inp_obj.keywords:
        if inp_obj.calctype == CalcType.conformer_search:
            toml_dict["runtype"] = "imtd-gc"
        elif inp_obj.calctype == CalcType.optimization:
            toml_dict["runtype"] = "optimize"
        elif inp_obj.calctype in {CalcType.energy, CalcType.gradient}:
            toml_dict["runtype"] = "singlepoint"
        elif inp_obj.calctype == CalcType.hessian:
            toml_dict["runtype"] = "numhess"
        else:
            raise EncoderError(
                f"Unsupported calctype {inp_obj.calctype} for CREST encoder.",
            )

    # Calculation level keywords
    calculation = toml_dict.pop("calculation", {})
    calculation_level = calculation.pop("level", [])
    if len(calculation_level) == 0:
        calculation_level.append({})
    for level_dict in calculation_level:
        level_dict["method"] = inp_obj.model.method
        level_dict["charge"] = inp_obj.structure.charge
        level_dict["uhf"] = inp_obj.structure.multiplicity - 1

    calculation["level"] = calculation_level
    toml_dict["calculation"] = calculation

    return toml_dict



import pytest

from qccodec.codec import decode, encode
from qccodec.encoders import terachem
from qccodec.exceptions import EncoderError


def test_main_terachem_energy(terachem_file):
    """Test the main terachem energy encoder."""
    contents = terachem_file("water.energy.out")
    computed_props = decode("terachem", "energy", stdout=contents)
    assert computed_props.energy == -76.3861099088


def test_encode_raises_error_with_invalid_calctype(prog_inp):
    prog_inp = prog_inp("transition_state")  # Not currently supported by crest encoder
    with pytest.raises(EncoderError):
        encode(prog_inp, "crest")


def test_main_terachem_encoder(prog_inp):
    prog_inp = prog_inp("energy")
    native_input = encode(prog_inp, "terachem")
    correct_tcin = (
        f"{'run':<{terachem.PADDING}} {prog_inp.calctype.value}\n"
        f"{'coordinates':<{terachem.PADDING}} {terachem.XYZ_FILENAME}\n"
        f"{'charge':<{terachem.PADDING}} {prog_inp.structure.charge}\n"
        f"{'spinmult':<{terachem.PADDING}} {prog_inp.structure.multiplicity}\n"
        f"{'method':<{terachem.PADDING}} {prog_inp.model.method}\n"
        f"{'basis':<{terachem.PADDING}} {prog_inp.model.basis}\n"
        f"{'purify':<{terachem.PADDING}} {prog_inp.keywords['purify']}\n"
        f"{'some-bool':<{terachem.PADDING}} "
        f"{str(prog_inp.keywords['some-bool']).lower()}\n"
    )
    assert native_input.input_file == correct_tcin



import math
import re
from enum import Enum
from pathlib import Path
from typing import Any, Generator, Optional, Union

import numpy as np
from qcio import (
    CalcType,
    ProgramInput,
    ProgramOutput,
    Provenance,
    SinglePointResults,
    Structure,
    constants,
)

from qccodec.exceptions import ParserError

from ..registry import register
from .utils import re_finditer, re_search


class CrestFileType(str, Enum):
    """CREST filetypes.

    Maps file types to their names as written in the CREST output directory (except
    for STDOUT and DIRECTORY).
    """

    STDOUT = "stdout"
    DIRECTORY = "directory"
    NUMHESS = "numhess1"
    G98 = "g98.out"
    ENGRAD = "crest.engrad"
    OPTLOG = "crestopt.log"


def iter_files(
    stdout: Optional[str], directory: Optional[Union[Path, str]]
) -> Generator[tuple[CrestFileType, Union[str, bytes, Path]], None, None]:
    """
    Iterate over the files in a CREST output directory.

    If stdout is provided, yields a tuple for it.

    If directory is provided, iterates over the directory to yield files according to
    program-specific logic.

    Args:
        stdout: The contents of the CREST stdout file.
        directory: The path to the directory containing the CREST output files.

    Yields:
        (FileType, contents) tuples for a program's output.
    """
    directory = Path(directory) if directory else None

    if stdout is not None:
        yield CrestFileType.STDOUT, stdout

    if directory is not None:
        # Check if the directory exists and is a directory
        if not directory.exists() or not directory.is_dir():
            raise ParserError(
                f"Directory {directory} does not exist or is not a directory."
            )
        yield CrestFileType.DIRECTORY, directory

        # Iterate over the files in the directory and yield their contents
        for filetype in CrestFileType:
            # Ignore STDOUT and DIRECTORY as they are handled above
            if filetype not in (CrestFileType.STDOUT, CrestFileType.DIRECTORY):
                # Check if the file exists in the directory
                file_path = directory / filetype.value
                if file_path.exists():
                    yield filetype, file_path.read_text()


@register(filetype=CrestFileType.STDOUT, target=("extras", "program_version"))
def parse_version(string: str) -> str:
    """Parse version string from CREST stdout.

    Matches format of 'crest --version' on command line.
    """
    regex = r"Version (\d+\.\d+\.\d+),"
    match = re_search(regex, string)
    return match.group(1)


@register(filetype=CrestFileType.DIRECTORY, calctypes=[CalcType.conformer_search])
def parse_conformers(
    directory: Union[Path, str], stdout: Optional[str], input_data: ProgramInput
) -> dict[str, Any]:
    """Parse the conformers from the output directory of a CREST conformer search calculation.

    Args:
        stdout: The contents of the CREST stdout file (not used).
        directory: Path to the directory containing the CREST output files.
        input_data: The input object used for the calculation

    Returns:
        The parsed conformers and their energies as a dictionary.
    """
    directory = Path(directory)
    conformers = Structure.open_multi(
        directory / "crest_conformers.xyz",
        charge=input_data.structure.charge,
        multiplicity=input_data.structure.multiplicity,
    )

    # CREST places the energy as the only value in the comment line
    conf_energies = [
        float(conf.extras[Structure._xyz_comment_key][0]) for conf in conformers
    ]

    # Add identifiers to the conformers if topo is unchanged
    _add_identifiers_to_structures(conformers, input_data)

    return {
        "conformers": conformers,
        "conformer_energies": conf_energies,
    }


@register(filetype=CrestFileType.DIRECTORY, calctypes=[CalcType.conformer_search])
def parse_rotamers(
    directory: Union[Path, str], stdout: Optional[str], input_data: ProgramInput
) -> dict[str, Any]:
    """Parse the rotamers from the output directory of a CREST conformer search calculation.

    Args:
        stdout: The contents of the CREST stdout file (not used).
        directory: Path to the directory containing the CREST output files.
        input_data: The input object used for the calculation

    Returns:
        The parsed rotamers and their energies as a dictionary.
    """
    directory = Path(directory)
    rotamers = Structure.open_multi(
        directory / "crest_rotamers.xyz",
        charge=input_data.structure.charge,
        multiplicity=input_data.structure.multiplicity,
    )

    # CREST places the energy as the only value in the comment line
    conf_energies = [
        float(conf.extras[Structure._xyz_comment_key][0]) for conf in rotamers
    ]

    # Add identifiers to the rotamers if topo is unchanged
    _add_identifiers_to_structures(rotamers, input_data)

    return {
        "rotamers": rotamers,
        "rotamer_energies": conf_energies,
    }


def _add_identifiers_to_structures(
    structures: list[Structure], input_data: ProgramInput
) -> None:
    """Add identifiers to a list of Structure objects if the 'topo' is unchanged."""
    if input_data.keywords.get("topo", True):
        ids = input_data.structure.identifiers.model_dump()
        for struct in structures:
            struct.add_identifiers(**ids)


@register(
    filetype=CrestFileType.ENGRAD,
    calctypes=[CalcType.energy, CalcType.gradient],
    target="energy",
)
def parse_energy(contents: str) -> float:
    """Parse the output of a CREST energy and gradient calculation.

    Args:
        contents: The text of the output file.

    Returns:
        The parsed energy as a float.
    """
    energy_regex = r"# Energy \( Eh \)\n#*\n\s*([-\d.]+)"
    return float(re_search(energy_regex, contents).group(1))


@register(
    filetype=CrestFileType.ENGRAD,
    calctypes=[CalcType.energy, CalcType.gradient],
    target="gradient",
)
def parse_gradient(contents: str) -> list[list[float]]:
    """Parse the output of a CREST energy and gradient calculation.

    Args:
        contents: The text of the output file.

    Returns:
        The parsed gradient as a Nx3 list of lists of floats.
    """
    gradient_regex = r"# Gradient \( Eh/a0 \)\n#\s*\n((?:\s*[-\d.]+\n)+)"
    vals = [float(x) for x in re_search(gradient_regex, contents).group(1).split()]
    # Group the values into chunks of 3 (for x, y, z).
    return [vals[i : i + 3] for i in range(0, len(vals), 3)]


@register(filetype=CrestFileType.STDOUT, calctypes=[CalcType.hessian], target="energy")
def parse_energy_numhess(contents: str) -> float:
    """Parse the initial singlepoint calculation energy from stdout.

    Args:
        contents: The text of the stdout file.

    Returns:
        The parsed energy.
    """
    energy_regex = r"Energy\s=\s*([-+]?\d+\.\d+)\s*Eh"
    return float(re_search(energy_regex, contents).group(1))


@register(
    filetype=CrestFileType.NUMHESS, calctypes=[CalcType.hessian], target="hessian"
)
def parse_numhess1(contents: str) -> list[list[float]]:
    """Parse the numerical Hessian from the CREST numhess1 file.

    Args:
        contents: The text of the numhess1 file.

    Returns:
        The parsed Hessian as a list of lists of floats.
    """
    float_regex = r"[-]?\d*\.\d+|\d+"
    numbers = [float(match.group()) for match in re_finditer(float_regex, contents)]
    sqrt_n = int(math.sqrt(len(numbers)))
    if sqrt_n * sqrt_n != len(numbers):
        raise ParserError(
            f"Expected a square matrix, but found {len(numbers)} elements."
        )
    # Reshape the flat list into a square matrix
    return [[numbers[i * sqrt_n + j] for j in range(sqrt_n)] for i in range(sqrt_n)]


@register(
    filetype=CrestFileType.G98, calctypes=[CalcType.hessian], target="freqs_wavenumber"
)
def parse_g98_freqs(contents: str) -> list[float]:
    """Parse the frequencies (wavenumbers) from G98 output text.

    Args:
        contents: The text of the g98 output file.

    Returns:
        A list of frequencies (wavenumbers) as floats.
    """
    regex = r"Frequencies\s+--\s+(?P<floats>(?:-?\d+\.\d+\s*)+)"
    # matches ['-335.2821                75.3406                87.4971\n ', ...]
    matches = re_finditer(regex, contents)
    return [float(freq) for match in matches for freq in match.group("floats").split()]


@register(
    filetype=CrestFileType.G98,
    calctypes=[CalcType.hessian],
    target="normal_modes_cartesian",
)
def parse_g98_normal_modes(contents: str) -> list[list[list[float]]]:
    """Parse the normal mode displacements from G98 output text.

    Args:
        contents: The text of the G98 output file.

    Returns:
        An (n_modes, n_atoms, 3) NumPy array representing the normal mode displacements
        with cartesian coordinates in Bohr.
    """
    block_re = re.compile(r"(Frequencies\s+--.*?)(?=Frequencies\s+--|$)", re.DOTALL)
    floats_re = re.compile(r"[-+]?\d*\.\d+")
    normal_modes_cartesian = []

    # Find all blocks of frequencies and displacements
    for block in block_re.findall(contents):
        # Determine the number of frequencies in this block from freqs line.
        n_freqs = len(block.splitlines()[0].split()[2:])
        # Match all floats in the block after "Atom AN"
        displacements = floats_re.findall(block.split("Atom AN")[1])
        # Initialize an empty list for each mode.
        mode_disp: list[list[list[float]]] = [[] for _ in range(n_freqs)]
        n_atoms = len(displacements) // (3 * n_freqs)
        for i in range(n_atoms):
            for j in range(n_freqs):
                index = i * (3 * n_freqs) + j * 3
                # Convert the displacement values to floats and scale to Bohr
                coords = [
                    float(val) * constants.ANGSTROM_TO_BOHR
                    for val in displacements[index : index + 3]
                ]
                mode_disp[j].append(coords)
        normal_modes_cartesian.extend(mode_disp)

    return normal_modes_cartesian


@register(
    filetype=CrestFileType.DIRECTORY,
    calctypes=[CalcType.optimization],
    target="trajectory",
)
def parse_trajectory(
    directory: Union[Path, str],
    stdout: str,
    input_data: ProgramInput,
) -> list[ProgramOutput]:
    """Parse the output directory of a CREST optimization calculation.

    Args:
        directory: Path to the directory containing the CREST output files.
        stdout: The stdout from CREST.
        input_data: The input object used for the calculation.

    Returns:
        The parsed optimization results as a list of ProgramOutput objects.
    """
    # Read in the xyz file containing the trajectory
    directory = Path(directory)
    xyz_text = (directory / "crestopt.log").read_text()

    # Parse structures and energies from the xyz file
    structures = Structure.from_xyz_multi(
        xyz_text,
        charge=input_data.structure.charge,
        multiplicity=input_data.structure.multiplicity,
    )
    energies = [
        float(struct.extras[Structure._xyz_comment_key][1]) for struct in structures
    ]

    # Fake gradient for each step because CREST does not output it
    fake_gradient = np.zeros(len(input_data.structure.symbols) * 3)

    # Parse program version
    program_version = parse_version(stdout)

    # Create the optimization trajectory
    trajectory: list[ProgramOutput] = [
        ProgramOutput(
            input_data=ProgramInput(
                calctype=CalcType.gradient,
                structure=struct,
                model=input_data.model,
            ),
            success=True,
            results=SinglePointResults(energy=energy, gradient=fake_gradient),
            provenance=Provenance(
                program="crest",
                program_version=program_version,
            ),
        )
        for struct, energy in zip(structures, energies)
    ]

    # Collect final gradient if calculation succeeded
    enegrad = directory / CrestFileType.ENGRAD.value
    if enegrad.exists():
        # Parse the energy and gradient from the file
        contents = enegrad.read_text()
        gradient = parse_gradient(contents)
        # Fill in final gradient
        trajectory[-1].results.gradient[:] = gradient

    else:
        # Calculation failed, so set the last .success = False
        final_po = trajectory[-1].model_dump()
        final_po["success"] = False
        trajectory[-1] = ProgramOutput(**final_po)

    return trajectory



import inspect
from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum
from typing import Callable, Optional, Union

from qcio import CalcType

from qccodec.exceptions import RegistryError


@dataclass
class ParserSpec:
    """Specification for a parser function.

    This class holds information about the parser function, the filetype it operates on,
    whether it is required, the calculation types it works on, and where to place the
    parsed output on the data collector.

    Attributes:
        parser: The parser function.
        filetype: The filetype that the parser is for.
        required: Whether a MatchNotFoundError should be raised in the decode function
            if the parser fails. Typically True, but should be set to False on parsers
            for values that are not always present such as excited state energies.
        calctypes: The calculation types that the parser works on.
        target: The location on the data collector where the parser's output should be placed.
        program: The name of the program that the parser is for. Set automatically by the
            `register` decorator.
    """

    parser: Callable
    filetype: Enum
    required: bool
    calctypes: list[CalcType]
    program: str
    target: Optional[Union[str, tuple[str, ...]]] = None

    def __post_init__(self):
        """Ensure that the parser function is Callable and that target exists unless filetype is directory."""
        if not callable(self.parser):
            raise RegistryError(f"Parser '{self.parser}' is not Callable.")
        # Ensure all non-directory parsers have a target
        if self.target is None and self.filetype.value != "directory":
            raise RegistryError(
                f"Parser '{self.parser.__name__}' for program "
                f"'{self.program}' must have a target if it is not a directory parser."
            )
        # Ensure that the target is a string or tuple of strings
        if self.target and not isinstance(self.target, (str, tuple)):
            raise RegistryError(
                f"Target '{self.target}' must be a string or tuple of strings."
            )


@dataclass
class ParserRegistry:
    """Registry for parser functions."""

    registry: dict[str, list[ParserSpec]] = field(
        default_factory=lambda: defaultdict(list)
    )

    def register(self, parser_spec: ParserSpec) -> None:
        """Register a new parser function. Enforce that no two parsers for the same program use the same target.

        Args:
            parser_spec: ParserSpec objects containing the parser function and
                information about the parser.
        """

        # Ensure that the target is unique for the program and filetype
        for registered_spec in self.registry.get(parser_spec.program, []):
            if (
                parser_spec.target is not None  # not directory parser
                and registered_spec.target == parser_spec.target  # Same target
                and set(parser_spec.calctypes)
                & set(registered_spec.calctypes)  # Shared
            ):
                raise RegistryError(
                    f"Duplicate parser target '{parser_spec.target}' and calctype "
                    f"'{set(parser_spec.calctypes)& set(registered_spec.calctypes)}' "
                    f"registered for program '{parser_spec.program}'."
                )
        self.registry[parser_spec.program].append(parser_spec)

    def get_parsers(
        self,
        program: str,
        filetype: Optional[Enum] = None,
        calctype: Optional[CalcType] = None,
    ) -> list[ParserSpec]:
        """Get all parser functions for a given program.

        Args:
            program: The program to get parsers for.
            filetype: Filter parsers for this filetype.
            calctype: Filter parsers for this calctype.

        Returns:
            List of ParserSpec objects.
        """

        parser_specs: list[ParserSpec] = self.registry[program]
        if not parser_specs:
            raise RegistryError(f"No parsers registered for program '{program}'.")

        # Filter parsers by filetype and calctype
        if filetype:
            parser_specs = [ps for ps in parser_specs if ps.filetype == filetype]

        if calctype:
            parser_specs = [ps for ps in parser_specs if calctype in ps.calctypes]
        return parser_specs

    def supported_programs(self) -> list[str]:
        """Get all programs with registered parsers.

        Returns:
            List of program names.
        """
        return list(self.registry.keys())

    def supported_filetypes(self, program: str) -> list[str]:
        """Get all filetypes for a given program.

        Args:
            program: The program to get filetypes for.

        Returns:
            List of filetypes.
        """
        return list(
            {parser_info.filetype.value for parser_info in self.get_parsers(program)}
        )

    def get_spec(self, parser: Callable) -> ParserSpec:
        """Get the ParserSpec for a given parser function.

        Args:
            parser: The parser function.

        Returns:
            The ParserSpec object.
        """
        for program, specs in self.registry.items():
            for spec in specs:
                if spec.parser == parser:
                    return spec
        raise RegistryError(f"No parser registered for '{parser.__name__}'.")


registry = ParserRegistry()


def register(
    *,
    filetype: Enum,
    calctypes: Optional[list[CalcType]] = None,
    required: bool = True,
    target: Optional[Union[str, tuple[str, ...]]] = None,
):
    """Decorator to register a function in the parser registry.

    Args:
        filetype: The filetype the parser operates on.
        calctypes: A list of calculation types on which this parser should operate. If None, it applies to all CalcTypes.
        required: If True and the parser fails a MatchNotFoundError will be raised.
            If False and the parser fails the value will be ignored.
        target: Where on the data collector to assign the parsed result.

    Example:
        ```python
        @register(filetype=FileType.stdout, calctypes=[CalcType.energy, CalcType.gradient], target="energy")
        def parse_energy(contents: str) -> float:
            # Parsing logic here
            return energy_value
        ```
    """

    def decorator(func):
        # Get the function's module name. Should match program name.
        module = inspect.getmodule(func).__name__
        program_name = module.split(".")[-1]

        # Create ParserSpec
        spec = ParserSpec(
            program=program_name,
            parser=func,
            filetype=filetype,
            required=required,
            # If calctypes is None, register for all CalcType values
            calctypes=calctypes if calctypes is not None else list(CalcType),
            target=target,
        )

        # Register the function in the global registry
        registry.register(spec)

        return func

    return decorator



import pytest

from qccodec.exceptions import RegistryError
from qccodec.registry import ParserSpec, registry


def test_get_parsers_program():
    parsers = registry.get_parsers("terachem")
    assert parsers
    for parser in parsers:
        assert isinstance(parser, ParserSpec)

    with pytest.raises(RegistryError):
        registry.get_parsers("not_a_program")


def test_get_parsers_program_filetype():
    parsers = registry.get_parsers("terachem", filetype="stdout")
    assert parsers
    for parser in parsers:
        assert isinstance(parser, ParserSpec)
        assert parser.filetype == "stdout"


def test_get_parsers_program_calctype():
    parsers = registry.get_parsers("terachem", calctype="gradient")
    assert parsers
    for parser in parsers:
        assert isinstance(parser, ParserSpec)
        assert "gradient" in parser.calctypes


def test_supported_programs():
    programs = registry.supported_programs()
    assert programs
    assert "terachem" in programs


def test_supported_filetypes():
    filetypes = registry.supported_filetypes("terachem")
    assert filetypes
    assert "stdout" in filetypes



from pathlib import Path

import pytest
from qcio import CalcType, ProgramInput

from qccodec.parsers.crest import (
    parse_conformers,
    parse_energy,
    parse_energy_numhess,
    parse_g98_freqs,
    parse_g98_normal_modes,
    parse_gradient,
    parse_numhess1,
    parse_rotamers,
    parse_trajectory,
    parse_version,
)

from .conftest import ParserTestCase, run_test_harness
from .data.crest.answers import (
    conformers,
    frequencies,
    gradients,
    hessians,
    normal_modes,
    optimizations,
)

######################################################
##### Top level tests for all registered parsers #####
######################################################


test_cases = [
    ParserTestCase(
        name="Parse version",
        parser=parse_version,
        contents=Path("crest_stdout.txt"),
        contents_stdout=True,
        calctype=CalcType.energy,
        success=True,
        answer="3.0.1",
    ),
    ParserTestCase(
        name="Parse version MatchNotFoundError",
        parser=parse_version,
        contents="No version to parse",
        contents_stdout=True,
        calctype=CalcType.energy,
        success=False,
    ),
    ParserTestCase(
        name="Parse energy CalcType.energy",
        parser=parse_energy,
        contents=Path("crest.engrad"),
        contents_stdout=False,
        calctype=CalcType.energy,
        success=True,
        answer=-0.335557824179335,
    ),
    ParserTestCase(
        name="Parse energy CalcType.gradient",
        parser=parse_energy,
        contents=Path("crest.engrad"),
        contents_stdout=False,
        calctype=CalcType.gradient,
        success=True,
        answer=-0.335557824179335,
    ),
    ParserTestCase(
        name="Parse gradient CalcType.gradient",
        parser=parse_gradient,
        contents=Path("crest.engrad"),
        contents_stdout=False,
        calctype=CalcType.gradient,
        success=True,
        answer=gradients.water,
    ),
    ParserTestCase(
        name="Parse gradient CalcType.energy",
        parser=parse_gradient,
        contents=Path("crest.engrad"),
        contents_stdout=False,
        calctype=CalcType.energy,
        success=True,
        answer=gradients.water,
    ),
    ParserTestCase(
        name="Parse energy CalcType.hessian",
        parser=parse_energy_numhess,
        contents=Path("hessian_stdout.txt"),
        contents_stdout=True,
        calctype=CalcType.hessian,
        success=True,
        answer=-110.490788960984773,
    ),
    ParserTestCase(
        name="Parse numhess1",
        parser=parse_numhess1,
        contents=Path("numhess1"),
        contents_stdout=False,
        calctype=CalcType.hessian,
        success=True,
        answer=hessians.water,
    ),
    ParserTestCase(
        name="Parse frequencies HF",
        parser=parse_g98_freqs,
        contents=Path("g98HF.out"),
        contents_stdout=False,
        calctype=CalcType.hessian,
        success=True,
        answer=frequencies.HF,
    ),
    ParserTestCase(
        name="Parse frequencies water",
        parser=parse_g98_freqs,
        contents=Path("g98.out"),
        contents_stdout=False,
        calctype=CalcType.hessian,
        success=True,
        answer=frequencies.water,
    ),
    ParserTestCase(
        name="Parse frequencies caffeine",
        parser=parse_g98_freqs,
        contents=Path("g98big.out"),
        contents_stdout=False,
        calctype=CalcType.hessian,
        success=True,
        answer=frequencies.caffeine,
    ),
    ParserTestCase(
        name="Parse normal_modes HF",
        parser=parse_g98_normal_modes,
        contents=Path("g98HF.out"),
        contents_stdout=False,
        calctype=CalcType.hessian,
        success=True,
        answer=normal_modes.HF,
    ),
    ParserTestCase(
        name="Parse normal_modes water",
        parser=parse_g98_normal_modes,
        contents=Path("g98.out"),
        contents_stdout=False,
        calctype=CalcType.hessian,
        success=True,
        answer=normal_modes.water,
    ),
    ParserTestCase(
        name="Parse normal_modes caffeine",
        parser=parse_g98_normal_modes,
        contents=Path("g98big.out"),
        contents_stdout=False,
        calctype=CalcType.hessian,
        success=True,
        answer=normal_modes.caffeine,
    ),
    ParserTestCase(
        name="Parse conformers",
        parser=parse_conformers,
        contents=Path("crest_conformers.xyz"),
        contents_stdout=False,
        calctype=CalcType.conformer_search,
        success=True,
        answer=conformers.conformers,
    ),
    ParserTestCase(
        name="Parse rotamers",
        parser=parse_rotamers,
        contents=Path("crest_rotamers.xyz"),
        contents_stdout=False,
        calctype=CalcType.conformer_search,
        success=True,
        answer=conformers.rotamers,
    ),
    ParserTestCase(
        name="Parse optimization",
        parser=parse_trajectory,
        contents=Path("optstdout.txt"),
        contents_stdout=True,
        calctype=CalcType.optimization,
        success=True,
        answer=optimizations.optimization,
        extra_files=["crest.engrad", "crestopt.log"],
    ),
]


@pytest.mark.parametrize("test_case", test_cases, ids=lambda tc: tc.name)
def test_crest_parsers(test_data_dir, prog_inp, tmp_path, test_case):
    """
    Tests the crest parsers to ensure that they correctly parse the output files and
    behave correctly within the decode function.
    """
    run_test_harness(test_data_dir, prog_inp, tmp_path, test_case)


####################################################
################ Additional Tests ##################
####################################################


def test_parse_conformers_charge_multiplicity_updates(test_data_dir, prog_inp):
    # Change charge and multiplicity in prog_input
    prog_input = prog_inp("conformer_search")
    prog_input_dict = prog_input.model_dump()
    prog_input_dict["structure"]["charge"] = -2
    prog_input_dict["structure"]["multiplicity"] = 3
    # Using fake prog_input for water; need .structure.charge and .multiplicity
    directory = test_data_dir / "crest"
    data = parse_conformers(directory, None, ProgramInput(**prog_input_dict))
    # Check conformer energies
    for struct in data["conformers"]:
        assert struct.charge == -2
        assert struct.multiplicity == 3


def test_parse_rotamers_charge_multiplicity_updates(test_data_dir, prog_inp):
    # Change charge and multiplicity in prog_input
    prog_input = prog_inp("conformer_search")
    prog_input_dict = prog_input.model_dump()
    prog_input_dict["structure"]["charge"] = 3
    prog_input_dict["structure"]["multiplicity"] = 5
    # Using fake prog_input for water; need .structure.charge and .multiplicity
    directory = test_data_dir / "crest"
    data = parse_rotamers(directory, None, ProgramInput(**prog_input_dict))
    # Check conformer energies
    for struct in data["rotamers"]:
        assert struct.charge == 3
        assert struct.multiplicity == 5



import subprocess

from qccodec.codec import decode


def test_cli(test_data_dir):
    # Call CLI script as a subprocess
    filepath = test_data_dir / "terachem" / "water.energy.out"
    result = subprocess.run(
        ["qccodec", "terachem", "energy", filepath], capture_output=True, text=True
    )
    # Check the return code
    assert result.returncode == 0

    # Check the output
    parse_result = decode("terachem", "energy", stdout=filepath.read_text())
    expected_output = parse_result.model_dump_json(indent=4, exclude_unset=True, exclude_defaults=True)
    assert result.stdout.strip() == expected_output



from importlib import metadata

__version__ = metadata.version(__name__)


from .codec import decode, encode  # noqa: F401
from .registry import registry  # noqa: F401

__all__ = ["decode", "encode", "registry"]



"""All parsers should follow a basic pattern:

1. Extract the parsed data, cast to its appropriate Python type, set it on the
    SinglePointResults object at the appropriate attribute name.
2. Raise a MatchNotFound error if a match was not found
3. Register parser with the registry by decorating it with the parser() decorator

Use the .utils.regex_search() helper function in place of re.search() to
ensure that a MatchNotFoundError will be raised in a parser. More sophisticated parsers
that use re.findall (like terachem.parse_hessian) or rely upon not finding a match may
implement a different interface, but please strive to follow this basic patterns as much
as possible.
"""

# Required for parsers to register
from .terachem import *  # noqa:  F403



import pytest

from qccodec.encoders.terachem import PADDING, XYZ_FILENAME, encode
from qccodec.exceptions import EncoderError


def test_write_input_files(prog_inp):
    """Test write_input_files method."""
    prog_inp = prog_inp("energy")

    native_input = encode(prog_inp)
    # Testing that we capture:
    # 1. Driver
    # 2. Structure
    # 3. Model
    # 4. Keywords (test booleans to lower case, ints, sts, floats)

    correct_tcin = (
        f"{'run':<{PADDING}} {prog_inp.calctype.value}\n"
        f"{'coordinates':<{PADDING}} {XYZ_FILENAME}\n"
        f"{'charge':<{PADDING}} {prog_inp.structure.charge}\n"
        f"{'spinmult':<{PADDING}} {prog_inp.structure.multiplicity}\n"
        f"{'method':<{PADDING}} {prog_inp.model.method}\n"
        f"{'basis':<{PADDING}} {prog_inp.model.basis}\n"
        f"{'purify':<{PADDING}} {prog_inp.keywords['purify']}\n"
        f"{'some-bool':<{PADDING}} "
        f"{str(prog_inp.keywords['some-bool']).lower()}\n"
    )
    assert native_input.input_file == correct_tcin


def test_write_input_files_renames_hessian_to_frequencies(prog_inp):
    """Test write_input_files method for hessian."""
    # Modify input to be a hessian calculation
    prog_inp = prog_inp("hessian")
    native_input = encode(prog_inp)

    assert native_input.input_file == (
        f"{'run':<{PADDING}} frequencies\n"
        f"{'coordinates':<{PADDING}} {XYZ_FILENAME}\n"
        f"{'charge':<{PADDING}} {prog_inp.structure.charge}\n"
        f"{'spinmult':<{PADDING}} {prog_inp.structure.multiplicity}\n"
        f"{'method':<{PADDING}} {prog_inp.model.method}\n"
        f"{'basis':<{PADDING}} {prog_inp.model.basis}\n"
        f"{'purify':<{PADDING}} {prog_inp.keywords['purify']}\n"
        f"{'some-bool':<{PADDING}} "
        f"{str(prog_inp.keywords['some-bool']).lower()}\n"
    )


def test_encode_raises_error_qcio_args_passes_as_keywords(prog_inp):
    """These keywords should not be in the .keywords dict. They belong on structured
    qcio objects instead."""
    qcio_keywords_from_terachem = ["charge", "spinmult", "method", "basis", "run"]
    prog_inp = prog_inp("energy")
    for keyword in qcio_keywords_from_terachem:
        prog_inp.keywords[keyword] = "some value"
        with pytest.raises(EncoderError):
            encode(prog_inp)



import pytest

from qccodec.exceptions import DataCollectorError
from qccodec.models import DataCollector


def test_data_collector_only_allows_one_value_per_key():
    dc = DataCollector()
    dc.add_data("energy", -74.964814)
    assert dc["energy"] == -74.964814

    with pytest.raises(DataCollectorError):
        dc.add_data("energy", -24.964814)
